/**
	@file nhdgen.c
	@author junnno(N.Inoue)
	@date 2002/12/08
	
	NHDwb_vO
*/

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>

#define SECTOR_LENGTH	512

#define MERGE_BUFFER_SIZE	(1024*1024*32)

#define NHD_FILE_ID	"T98HDDIMAGE.R0"

#pragma option -a1

typedef struct {
    char  szFileID[15];                 // ID "T98HDDIMAGE.R0"
    char  Reserve1[1];                  // \
    char  szComment[0x100];             // C[WRg(ASCIIz)
    DWORD dwHeadSize;                   // wb_̃TCY
    DWORD dwCylinder;                   // V_
    WORD  wHead;                        // wbh
    WORD  wSect;                        // PgbÑZN^
    WORD  wSectLen;                     // ZN^
    char  Reserve2[2];                  // \
    char  Reserve3[0xe0];               // \
}NHD_FILE_HEAD,*LP_NHD_FILE_HEAD;


typedef struct
{
	BYTE Jump[3];
	BYTE OEMName[8];
	WORD BytesPerSector;
	BYTE SectorsPerCluster;
	WORD ReservedSector;
	BYTE FATCount;
	WORD RootDirEntriesCount;
	WORD SectorsCount;
	BYTE MediaType;
	WORD FATSectors;
	WORD BIOSSectors;
	WORD BIOSHeads;
	DWORD HiddenSectors;
	DWORD DoubleSectorsCount;
	BYTE OSData[3];
	DWORD VolumeSerial;
	BYTE VolumeLabel[11];
	BYTE FileSystem[8];
} BOOT_RECORD_FAT16;

typedef struct
{
	BYTE Jump[3];
	BYTE OEMName[8];
	WORD BytesPerSector;
	BYTE SectorsPerCluster;
	WORD ReservedSector;
	BYTE FATCount;
	WORD RootDirEntriesCount;	// must be zero
	WORD SectorsCount;			// must be zero
	BYTE MediaType;
	WORD FATSectors;            // must be zero
	WORD BIOSSectors;
	WORD BIOSHeads;
	DWORD HiddenSectors;
	DWORD DoubleSectorsCount;

    // FAT32
	DWORD DoubleFATSectorsCount;
    WORD  wFlags;
    WORD  wVersion;
    DWORD dwRootDirStart;
    WORD  wInfoStart;
    WORD  wBackupStart;
    BYTE  Reserved[12];
    BYTE  cLogicalDriveNumber;
    BYTE  cUnused;
    BYTE  cExtendedSignature;
    DWORD dwSerialNumber;
    BYTE  strVolumeName[11];
    BYTE  strFATName[8];  		// must be "FAT32"
    BYTE  Program[420];
    BYTE  strSignature[2];
} BOOT_RECORD_FAT32;

#pragma option -a.


long getFileSize(FILE *fp)
{
	long current;
	long last;
	
	current = ftell(fp);
	fseek(fp, 0, SEEK_END);
	last = ftell(fp);
	fseek(fp, current, SEEK_SET);
	
	return last;
}

int autoHeader(const char* srcfile, LP_NHD_FILE_HEAD pnfh)
{
	int i;
	char buf[SECTOR_LENGTH];
	FILE* fp;
	
	fp = fopen(srcfile, "rb");
	if(!fp){
		printf("Source file '%s' cannot be opened.\n", srcfile);
		return 0;
	}
	
	for(i=0; !feof(fp); i++){
		if(i%100==0) printf("%d bytes was scanned.\n", i*SECTOR_LENGTH);
		fread(buf, SECTOR_LENGTH, 1, fp);
		if( (memcmp(((BOOT_RECORD_FAT16*)buf)->FileSystem, "FAT16   ", 8)==0)	// u[gubN̏B
          ||(memcmp(((BOOT_RECORD_FAT16*)buf)->FileSystem, "FAT12   ", 8)==0)	// ΏHDDDOS̃o[WÂƎs邩H
          ||(memcmp(((BOOT_RECORD_FAT32*)buf)->strFATName, "FAT32   ", 8)==0))
        {
			BOOT_RECORD_FAT16* pbr = (BOOT_RECORD_FAT16*)buf;
			printf("Partition boot block was found at offset %d.\n", ftell(fp));
			pnfh->wHead = pbr->BIOSHeads;
			pnfh->wSect = pbr->BIOSSectors;
			pnfh->wSectLen = SECTOR_LENGTH;
			pnfh->dwCylinder = getFileSize(fp)/(pnfh->wHead*pnfh->wSect*pnfh->wSectLen);
			printf("Disk parameter is...\nHeads: %u\nSectors: %u\nCylinders: %u\nSector length: %u\n\n",
				pnfh->wHead, pnfh->wSect, pnfh->dwCylinder, pnfh->wSectLen);
			fclose(fp);
			return 1;
		}
	}
	fclose(fp);
	printf("Patition boot block was not found.\n");
	return 0;
}

int manualHeader(LP_NHD_FILE_HEAD pnfh)
{
	char strbuf[0x100];
	DWORD size;
	
	printf("Please input number of heads.\n");
	gets(strbuf);
	pnfh->wHead = atoi(strbuf);
	printf("Input number of sectors.\n");
	gets(strbuf);
	pnfh->wSect = atoi(strbuf);
	printf("Input sector length. It is 512 bytes typically.\n");
	gets(strbuf);
	pnfh->wSectLen = atoi(strbuf);
	printf("Input number of cylinders.\n");
	gets(strbuf);
	pnfh->dwCylinder = atoi(strbuf);
	size = pnfh->wHead*pnfh->wSect*pnfh->wSectLen*pnfh->dwCylinder;
	printf("OK, Image file size is %u %03u %03u bytes.\n\n",
		size/1000000, size/1000%1000, size%1000);
	
	return 1;
}

int mergeHeader(LP_NHD_FILE_HEAD pnfh, const char *src, const char *dst)
{
	FILE *fps, *fpd;
	char *buf;
	int result = 1;
	
	buf = malloc(MERGE_BUFFER_SIZE);
	if(!buf){
		printf("It is failed to allocate memory.\n");
		return 0;
	}
	
	fps = fopen(src, "rb");
	if(fps){
		fpd = fopen(dst, "wb");
		if(fpd){
			int size;
			int i;
			
			fwrite(pnfh, sizeof(NHD_FILE_HEAD), 1, fpd);
			if(ferror(fpd)){
				printf("Header writing error.\n");
				result = 0;
			}
			else{
				printf("Header write ok.\n");
				for(i=0; !feof(fps);){
					printf("Reading...\n");
					size = fread(buf, 1, MERGE_BUFFER_SIZE, fps);
					if(ferror(fps)){
						printf("Data reading error.\n");
						result = 0;
						break;
					}
					printf("Writing...\n");
					fwrite(buf, size, 1, fpd);
					if(ferror(fpd)){
						printf("Data writing error.\n");
						result = 0;
						break;
					}
					fflush(fpd);
					i+=size;
					printf("%d bytes done.\n", i);
				}
			}
			fclose(fpd);
		}
		fclose(fps);
	}
	free(buf);
	return result;
}

int writeHeader(LP_NHD_FILE_HEAD pnfh, const char* dstfile)
{
	FILE* fp;
	
	printf("Writing...\n");
	fp = fopen(dstfile, "wb");
	if(fp){
		fwrite(pnfh, sizeof(NHD_FILE_HEAD), 1, fp);
		fclose(fp);
		printf("Done.\n");
		return 1;
	}
	else{
		printf("Sorry, it was failed to open the file '%s'. Every inputs were discarded.\n", dstfile);
		return 0;
	}
}

int main(int argc, char **argv)
{
	NHD_FILE_HEAD nfh;
	FILE* fp;
	char strbuf[0x100];
	int i;
	const char *srcfile = "";
	const char *dstfile = "";
	int manual = 1;
	int merge = 0;
	
	for(i=1; i<argc; i++){
		if(argv[i][0] == '/'){
			if(strcmp(argv[i], "/auto")==0) manual = 0;
			else if(strcmp(argv[i], "/merge")==0) merge = 1;
		}
		else{
			if(!srcfile[0] && (merge || !manual)) srcfile = argv[i];
			else if(!dstfile[0]) dstfile = argv[i];
		}
	}
	
	if( !dstfile[0] || ((merge || !manual) && !srcfile[0])){
		printf(
			"Usage: nhdgen [/auto] [/merge] C[Wt@C o̓t@C\n\n"
			"/auto  : [hBw肵Ȃꍇ͎蓮[hB\n"
			"/merge : wb_ƃfBXNC[WNHDt@Co͂B\n"
			"         w肵Ȃꍇ̓wb_݂̂o͂B\n"
			"蓮[hłȂꍇ͏o̓t@Cw肵܂B\n\n"
			") nhdgen /auto /merge hdd.bin hdd.nhd\n"
			"hdd.binp[^FAwb_hdd.nhdɏo͂܂B\n");
		return 0;
	}
	
	printf("Mode: %s\n", manual?"manual":"auto");
	printf("Merge: %s\n", merge?"yes":"no");
	printf("Source file: %s\n", srcfile[0]?srcfile:"(none)");
	printf("Destination file: %s\n", dstfile[0]?dstfile:"(none)");
	
	memset(&nfh, 0, sizeof(nfh));	// zero memory
	memcpy(nfh.szFileID, NHD_FILE_ID, 15);	// header signature
	nfh.dwHeadSize = sizeof(nfh);			// header size
	printf("Input the disk image comment. (255 letters max)\n");
	gets(strbuf);
	strncpy(nfh.szComment, strbuf, sizeof(nfh.szComment));
	
	if(manual){
		manualHeader(&nfh);
	}else{
		autoHeader(srcfile, &nfh);
	}
	
	if(merge){
		mergeHeader(&nfh, srcfile, dstfile);
	}else{
		writeHeader(&nfh, dstfile);
	}
	
	return 0;
}
